import asyncio
import functools
import collections

class AsyncCreditSemaphore():
    def __init__(self, credits=1):
        if credits < 0:
            raise ValueError("Semaphore credit value must be >= 0")
        self._waiters = None
        self._credits = credits
        self._future_costs = {}
        self._future_refunds = {}

    def __repr__(self):
        res = super().__repr__()
        return f"has credits {self._credits} and waiter count of {0 if not self._waiters else len(self._waiters)}."

    async def transact(self, coroutine, credits, refund_time, transaction_id=None, verbose=False):
        assert(asyncio.iscoroutine(coroutine))
        
        if verbose: 
            print(f"TXN {transaction_id} acquiring CreditSemaphore")
        await self.acquire(credits, refund_time)

        if verbose: 
            print(f'TXN {transaction_id} entered CreditSemaphore...')
        try:
            result = await coroutine
        except Exception as err:
            raise err
        finally:
            self.refund_later(credits, refund_time)
        
        if verbose: 
            print(f'TXN {transaction_id} exits CreditSemaphore, schedule refund in {refund_time}...')
        return result

    def refund_later(self, credits, after_time):
        assert(after_time >= 0)
        if after_time == 0:
            self.release(credits)
        else:
            asyncio.get_running_loop().call_later(
                after_time,
                functools.partial(self.release, credits)
            )
        return

    def locked(self, require_credits):
        return self._credits < require_credits or (
            any(not w.cancelled() for w in (self._waiters or ())))

    async def acquire(self, require_credits, refund_time):
        if not self.locked(require_credits):
            self._credits -= require_credits
            return True

        if self._waiters is None:
            self._waiters = collections.deque()
        
        fut = asyncio.get_event_loop().create_future()
        self._waiters.append(fut)
        self._future_costs[fut] = require_credits
        self._future_refunds[fut] = refund_time

        try:
            try:
                await fut
            finally:
                del self._future_costs[fut]
                del self._future_refunds[fut]
                self._waiters.remove(fut)

        except Exception as err:
            if not fut.cancelled():
                self.refund_later(self._future_costs[fut], self._future_refunds[fut])
                del self._future_costs[fut]
                del self._future_refunds[fut]
            raise

        self._wake_up_next()
        return True

    def release(self, taken_credits):
        self._credits += taken_credits
        self._wake_up_next()

    def _wake_up_next(self):
        if not self._waiters:
            return

        for fut in self._waiters:
            if not fut.done() and self._credits >= self._future_costs[fut]:
                self._credits -= self._future_costs[fut]
                fut.set_result(True)
                return


import asyncio
from functools import wraps

def consume_credits(_func=None, *, costs, refund_in, attrname="credit_semaphore", timeout=0, verbose=True):
    
    def sync_wrapper(func):
        @wraps(func)
        async def connection_handler(self, *args, **kwargs):    
            assert(hasattr(self, attrname))
            assert(costs >= 0)

            sem = getattr(self, attrname)
            transaction = sem.transact(
                coroutine=func(self, *args, **kwargs), 
                credits=costs, 
                refund_time=refund_in, 
                transaction_id={"fn": func.__name__, "args": args, "kwargs": kwargs}, 
                verbose=verbose
            )

            if timeout == 0:
                result = await transaction
            else:
                result = await asyncio.wait_for(asyncio.create_task(transaction), timeout=timeout)
  
            return result
                      
        return connection_handler

    if _func is None:
        return sync_wrapper
    else:
        return sync_wrapper(_func)


import setuptools
from os.path import join, dirname

with open(join(dirname(__file__), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name='credit_semaphore',
    version='0.1.1',
    description='Asynchronous Semaphore Based on Credits for Efficient Credit-Based API Throttling',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='HangukQuant',
    url="https://github.com/hangukquant/credit_semaphore",
    download_url="https://pypi.org/project/credit-semaphore/",
    install_requires=[
        'opencv-python'
    ],
    author_email='',
    packages=setuptools.find_packages()
)